////////////////////////////////////////////////////////////////////////////////
// Wonderswan emulator
////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <commctrl.h>
#include <direct.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <io.h>
#include <fcntl.h>
#include <conio.h>
#include <time.h>
#include "./Ptc/source/ptc.h"
#include "./source/log.h"
#include "./source/rom.h"
#include "./source/nec/nec.h"
#include "./source/memory.h"
#include "./source/gpu.h"
#include "./source/io.h"
#include "./source/ws.h"


////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////
#define LOG_PATH "wswan.log"

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////
void ws_emulate(void)
{
	#define KEY_ENTER	0x0D
	#define KEY_ESC		0x1b
	#define KEY_UP		0x26
	#define KEY_DOWN	0x28
	#define KEY_LEFT	0x25
	#define KEY_RIGHT	0x27
	#define KEY_BUTTON1 0x57
	#define KEY_BUTTON2 0x58

	uint32		 startTime, endTime, totalFrames;
	unsigned int nNormalLast=0;
	int			 nNormalFrac=0; 
	int			 nTime=0,nCount=0; int i=0;

	// 15 bits RGB555
	Format format(16,0x007c00,0x00003e0,0x0000001f);
	Console console;
	Surface surface(224,144,format);

	console.option("DirectX");
	console.option("windowed output");
	console.option("fixed window");
	console.option("center window");

	console.open("Oswan",224,144,format);
	int16 *backbuffer=(int16*)malloc(224*144*sizeof(int16));
	memset(backbuffer,0x00,224*144*sizeof(int16));
	
	totalFrames=0;
	startTime=clock();
	nNormalLast=0;// Last value of timeGetTime()
	nNormalFrac=0; // Extra fraction we did
	nNormalLast=timeGetTime();
	while (1)
	{
		nTime=timeGetTime()-nNormalLast;					  // calcule le temps coul depuis le dernier affichage
															  // nTime est en mili-secondes.
		// dtermine le nombre de trames  passer + 1
		nCount=(nTime*600 - nNormalFrac) /10000; 	

		// si le nombre de trames  passer + 1 est nul ou ngatif,
		// ne rien faire pendant 2 ms
		if (nCount<=0) 
		{ 
			Sleep(2); 
		} // No need to do anything for a bit
		else
		{
			nNormalFrac+=nCount*10000;				// 
			nNormalLast+=nNormalFrac/600;				// add the duration of nNormalFrac frames
			nNormalFrac%=600;							// 

			// Pas plus de 9 (10-1) trames non affiches 
			if (nCount>10) 
			  nCount=10; 

			ws_key_start=0;
			ws_key_left=0;
			ws_key_right=0;
			ws_key_up=0;
			ws_key_down=0;
			ws_key_button_1=0;
			ws_key_button_2=0;
			int ws_key_esc=0;

			while (console.key())
			{

				Key k=console.read();

				if (k.code()==KEY_ESC) ws_key_esc=1;
				
				switch (k.code())
				{
				case KEY_ENTER:		ws_key_start=1;		break;
				case KEY_UP:		ws_key_up=1;		break;
				case KEY_DOWN:		ws_key_down=1;		break;
				case KEY_RIGHT:		ws_key_right=1;		break;
				case KEY_LEFT:		ws_key_left=1;		break;
				case KEY_BUTTON1:	ws_key_button_1=1;	break;
				case KEY_BUTTON2:	ws_key_button_2=1;	break;
				}
			}
			if (ws_key_esc) break;

			for (i=0;i<nCount-1;i++) 
				while (!ws_executeLine(backbuffer,0));
		
			while (!ws_executeLine(backbuffer,1));
			totalFrames++;
			int32 *vs = (int32 *)surface.lock();
			memcpy(vs,backbuffer,224*144*2);
			surface.unlock();
			surface.copy(console);
			console.update();
		}
	}
	endTime=clock();
	float fps=totalFrames/(((float)(endTime-startTime))/(float)CLOCKS_PER_SEC);
	printf("%f fps (%i %% the original speed)\n",fps, (int)((fps*100)/60));
	console.close();
}
////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////
char *gui_getRomPath(void)
{
	InitCommonControls();
	static OPENFILENAME ofn;       // common dialog box structure
	char szFile[260];       // buffer for file name
	HWND hwnd;              // owner window
	HANDLE hf;              // file handle
	static char oldDir[1024];

	_getcwd(oldDir,1024);
	// Initialize OPENFILENAME
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = NULL;
	ofn.lpstrFile = szFile;
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrFilter = "Wonderswan mono\0*.ws\0Wonderswan color\0*.wsc\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
	ofn.hInstance=NULL;
	// Display the Open dialog box. 

	if (GetOpenFileName(&ofn)==TRUE)
		return(ofn.lpstrFile);
	chdir(oldDir);
	return(NULL);
}
////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////
int main(int nbArgs, char *args[])
{
	printf("Wonderswan emulator v0.50 (last build on %s %s)\n",__DATE__,__TIME__);
	
	if (!log_init(LOG_PATH)) printf("Warning: cannot open log file %s\n",LOG_PATH);
	while (1)
	{
		
		char *rom_path;
		if (nbArgs!=2)
			rom_path=gui_getRomPath();
		else
			rom_path=args[1];

		if (rom_path==NULL)
			break;

		if (!ws_init(rom_path))
		{
			printf("Error: %s cannot be loaded\n",rom_path);
			return(0);
		}
		ws_reset();
		ws_emulate();
		ws_done();
		break;
	}

	log_done();
	return(0);
}						